//                              -*- Mode: C++ -*-
//: EALinkAnalyzer.C
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 13/8/97
// Last Update ........
//
#include "care/care.hh"
#include "care/EMCluster.hh"
#include "care/EALinkAnalyzer.hh"
#include "arte/ArteMessage.hh"
#include "arte/ArteProto.hh"
#include "care/EMTrackOnEcal.hh"
#include "care/MMBank.hh"
#define MCARLO
#ifdef MCARLO
#  include "care/EAMCTruthAnalyzer.hh"
#endif
#undef WRITE
#undef DEBUG

#include "auto/RCLX.hh"
/* 
CLASS
  EALinkAnalyzer
  Class for defining and updating the links between the
  RCCL bank and the banks HITC and RTRA.

*/

//////////////////// EALinkAnalyzer Class Definitions ////////////////

// No functions
EALinkAnalyzer::EALinkAnalyzer() : mkHitRel(0), mkTraRel(0), nrccl(0) {
  // initialize data for the relational bank
}

EALinkAnalyzer::~EALinkAnalyzer(){}


//--
// Cluster Analyzer
void EALinkAnalyzer::doAnalysis(const EMCluster * cluster){
  // loop on all cluster members in REVERSE directions since in this way the
  // relations are stored in DIRECT order...
  for(const EMClusterMember* child = cluster->rbegin(); 
      child!=cluster->rend();  child = child->previous())
    child->analyzeMe(*this);
}

//--
// ClusterTree Analyzer
void EALinkAnalyzer::doAnalysis(const EMClusterTree * ){}


//--
// HitCell Analyzer
void EALinkAnalyzer::doAnalysis(const EMHitCell *hitC){
# ifndef ONLINE
  if(mkHitRel){
    unsigned int i = hitC->whichBank().index();
    // write the relation for this cell
    if( i> ArtePointer<HITC>::nrow() || i==0 || 
	nrccl< 1|| nrccl>ArtePointer<RCCL>::nrow() ){
      arteMsg<<" Error: hitc: "<<i<<"  RCCL: "<<nrccl;
      arteMsg.warn();
      arteMsg<<"        hitc: "<< ArtePointer<HITC>::nrow()
	  <<"  RCCL: "<<ArtePointer<RCCL>::nrow();
      arteMsg.warn();
      hitC->getParent()->dump();
    }
    MMBank<RCLX_HITC>::Brel(i, nrccl);
  }
# endif
}
//--
//  TrackOnEcal Analyzer
void EALinkAnalyzer::doAnalysis(const EMTrackOnEcal * track){
# ifndef ONLINE
  if(mkTraRel){
    // get the track index number
#   ifdef CMATCH
    unsigned int i = track->Infos()->link().index();
#   else
    unsigned int i = track->Infos().track.index();
#   endif
#   ifdef RTRALINK
    if( i> ArtePointer<RTRA>::nrow() || i==0 || 
	nrccl< 1|| nrccl>ArtePointer<RCCL>::nrow()){
      arteMsg<<" Error: track: "<<i<<"  RCCL: "<<nrccl;
      arteMsg.warn();
      arteMsg<<"        rtra: "<< ArtePointer<RTRA>::nrow()
	  <<"  RCCL: "<<ArtePointer<RCCL>::nrow();
      arteMsg.warn();
      track->getParent()->dump();
    }
    // write the relation for this cell
    MMBank<RCLX_RTRA>::Brel(i, nrccl);
#  else
    if( i> ArtePointer<RSEG>::nrow() || i==0 || 
	nrccl< 1|| nrccl>ArtePointer<RCCL>::nrow()){
      arteMsg<<" Error: track: "<<i<<"  RCCL: "<<nrccl;
      arteMsg.warn();
      arteMsg<<"        rseg: "<< ArtePointer<RSEG>::nrow()
	     <<"  RCCL: "<<ArtePointer<RCCL>::nrow();
      arteMsg.warn();
      track->getParent()->dump();
    }
    // write the relation for this cell
    MMBank<RCLX_RSEG>::Brel(i, nrccl);
#  endif
  }
# endif
}

//--
// EMClusterMember Analyzer
void EALinkAnalyzer::doAnalysis(const EMClusterMember *){}

//--
//  Fill totally new relations
//
void EALinkAnalyzer::fillNewEntry(const EMCluster * cluster, 
				  ArtePointer<RCCL> rccl){
  // now build relational bank RCLX_HITC and RCLX_RTRA
  mkHitRel = true;
  mkTraRel = true;
  nrccl = rccl.index();
  doAnalysis(cluster);
  mkHitRel = false;
  mkTraRel = false;

# ifdef WRITE
  testLinks(cluster, rccl);
# endif

# ifdef MCARLO
  EAMCTruthAnalyzer mct;
  mct.reset();
  mct.doAnalysis(cluster);
  mct.sortDeposits();
  ECMCDeposit  dep = mct.maxDeposit();
  // Evaluate the total deposit in a cell
  Energy_t edep = 0.0;
  vector<ECMCDeposit>::const_iterator first = mct.getVector().begin();
  vector<ECMCDeposit>::const_iterator last = mct.getVector().end();
  for(vector<ECMCDeposit>::const_iterator iter = first; iter!=last; iter++){
    edep += iter->energy();
  }
  // Evaluate the 80 % of the deposit
  Energy_t edep1 = edep* 0.80;
  vector<ECMCDeposit>::const_iterator iter1 = first;
  for(; iter1!=last && edep1>0.0; iter1++){
    edep1 -= iter1->energy();
  }
  last=iter1;

  vector<ECMCDeposit>::const_reverse_iterator rvfirst = 
	 mct.getVector().rend();
  for( vector<ECMCDeposit>::const_reverse_iterator rviter1 = 
	  mct.getVector().rbegin(); 
       rviter1!=rvfirst; rviter1++){
    
    // get the track index number
    unsigned int i = (*rviter1).track().index();
    if( i> ArtePointer<MTRA>::nrow() || 
	nrccl<1 || nrccl>ArtePointer<RCCL>::nrow()){
      arteMsg<<" Error: MC track: "<<i<<"  RCCL: "<<nrccl;
      arteMsg.warn();
      arteMsg<<"        mtra: "<< ArtePointer<MTRA>::nrow()
	     <<"  RCCL: "<<ArtePointer<RCCL>::nrow();
      arteMsg.warn();
      cluster->dump();
    } else if( i!=0 ) {
      MMBank<RCLX_MTRA>::Brel(i, nrccl);
    }
  }
# endif
}

//--
//  Evaluate the information for the cluster 
//  and fill the TRACK relation bank only
//

void EALinkAnalyzer::updateTrackRelEntry(const EMCluster * cluster, 
					 ArtePointer<RCCL> rccl){
  // now build relational bank RCLX_RTRA
  mkHitRel = false;
  mkTraRel = true;
  nrccl = rccl.index();
  doAnalysis(cluster);
  mkHitRel = false;
  mkTraRel = false;

# ifdef WRITE
  testLinks(cluster, rccl);
# endif
}

//--
//  Evaluate the information for the cluster and 
//  fill an old entry in the bank
//
void EALinkAnalyzer::updateOldEntry(const EMCluster * , 
				    ArtePointer<RCCL> ){}

//--
//  Test the links
//
#ifndef DEBUG
void EALinkAnalyzer::testLinks(const EMCluster * , ArtePointer<RCCL> ){
 //
#else
void EALinkAnalyzer::testLinks(const EMCluster * clus, 
			       ArtePointer<RCCL> rccl){
 //
  arteMsg<<"Testing cluster";
  arteMsg.allp();
  clus->dump();
  // test the link done
  arteMsg<<"Cluster number:"<<rccl.index()
      <<" at "<<rccl->x<<","<<rccl->y
      <<" with "<<rccl->wid<<" cells and "<<rccl->esum<<" GeV";
  arteMsg.allp();
  // now loop on relations of this RCCL with HITC
  ArtePointer<RCLX> rclx(rccl);
  for( ArteRelationIterator<RCLX_HITC> relIter(rclx->hitc);
       relIter.valid(); relIter.next()){
    ArtePointer<HITC> cell = ArtePointer<HITC>(relIter);
    //    cout.form("->Cell: ID=%6d  Energy = %7.3f  Pos:(%7.2f, %7.2f ) ",
    //        cell->gcce,cell->q,cell->x,cell->y);
    arteMsg<<"->Cell: ID="<< cell->gcce<<"  Energy = "<<cell->q<<"  Pos:("
	<<cell->x<<", "<<cell->y<<") ";
    arteMsg.allp();
  }
  // now loop on relations of this RCCL with RTRA
  for( ArteRelationIterator<RCLX_RTRA> relIterat(rclx->rtra);
       relIterat.valid(); relIterat.next()){
    ArtePointer<RTRA> rtra = ArtePointer<RTRA>(relIterat);
    arteMsg<<"->Track: ID="<<rtra<<"  Momentum = "<< 1./rtra->pf;
    arteMsg.allp();
  }
#endif
}

