//                              -*- Mode: C++ -*-
//:ECALDetector.C
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 1/10/97
// Last Update ........
//
//

#include "care/ECALDetector.hh"
#include "care/care.hh"
#include "care/ECBuilder.hh"
#include "care/EMClusterTree.hh"
#ifdef PERFORMANCES
#  include "arte/Chrono.hh"
#  include "arte/ChronoManager.hh"
#endif
#include "arte/ArteMessage.hh"
#include "care/caredata.hh"
#include "care/ELLUTsManager.hh"
#include "care/ECHitCellArray.hh"
#include "care/EDClusHypotModifier.hh"
#include "care/EDClusTraHypotModifier.hh"
#include "care/EDBoundCellModifier.hh"
#include "care/EAHypTraAnalyzer.hh"
#include "care/EAHypAnalyzer.hh"
#include "care/ECMCcalib.hh"
#include "care/CmdCareDump.hh"
#include "care/CmdCareLevel.hh"
#include "care/MMBank.hh"
#ifndef ONLINE
#include "auto/RTRA.hh"
#endif

extern "C" void dgcalinit_();

const char VERSION[] = "2.00/00";

#ifdef PERFORMANCES

Chrono *  filling = 0;
Chrono *  init = 0;
Chrono *  clustering = 0;
Chrono *  merging = 0;
Chrono *  bankwrite = 0;
Chrono *  firststage = 0;
Chrono *  secondstage = 0;
Chrono *  matching = 0;
Chrono *  matching1 = 0;
Chrono *  extrap = 0;
Chrono *  matching2 = 0;
Chrono *  matching3 = 0;
Chrono *  splitting = 0;
Chrono *  bankupdate = 0;

#endif

static char stars[] = 
      "***********************************************************";

#undef WRITE


/**
class ECALDetector.
  Provide a collection of basic function for the ECAL detector 
  reconstruction and analysis.

*/

ECALDetector* ECALDetector::ecalDetector = 0;

ECALDetector::ECALDetector(){
  topClusterTree = new EMClusterTree();
  clusBuilder = new ECBuilder();
  pSignalCells = new ECHitCellArray();
  int* storage = new int[5000000]; // reserve 4 mega
  delete [] storage;
}

void ECALDetector::initialize() {
  if(ecalDetector==0){
    ecalDetector = new ECALDetector();

    if( !ECalInfo.init){
      //
      // Searching for Calorimeter informations in geometry tables
      //
      // Ecal Pointer to GESL table
      ArteTableIterator<GESL> ecalp;
      for(; ecalp.valid(); ecalp.next()){
	if(!strncmp(ecalp->name,"CAL",3)){
	  // ECAL found. ( it could be CALL CALS CALW)
	  // Exit this loop
	  break;
	}
      }
      if( ecalp.valid() ){
	//
	// Notify calorimeter initialization
	//
	arteMsg<<stars;
	arteMsg.allp();
	arteMsg<<" Care  "<<VERSION;
	arteMsg.allp();
	arteMsg<<" Monte Carlo Version ";
	arteMsg.allp();
	arteMsg<<stars;
	arteMsg.allp();
	
	//
	// Initialize the LUTs  
	ELLUTsManager::instance();  
	//
#       ifdef PERFORMANCES
          // Initialize the ChronoManager
          ChronoManager::instance();
#       endif
	//
	dgcalinit_();
	//
	arteMsg<<stars;
	arteMsg.allp();  
	// 
	
	ECMCcalib::dumpCalib();    // printout calibration data
	ECHitCellArray::dumpCutOff();   // printout cutoff data
	//
	EDClusHypotModifier::dumpMergingCutOffs(); // printout of merging data
	arteMsg<<stars;
	arteMsg.allp();
	// 
	// set to true the initialization flag
	ECalInfo.init = true;
      } else {
	//
	//  ECAL not found: print a warning
	//
	arteMsg<<"ECAL tables not found: "
	       <<"Calorimeter reconstruction unavailable";
	arteMsg.warn();
      }
    } else {
      //
      //  careinit called more than once
      //
      arteMsg<<"More than one call to careinit_() :  check your code !";
      arteMsg.error();
    }
  }
}

//--
/// Find and return shower hypotheses
const vector<const EMCluster*> & ECALDetector::findShowers(){
  EcalEvent_t * event=0;
  return findShowers(event);
}

//--
/// Find and return shower hypotheses
const vector<const EMCluster*> & ECALDetector::findShowers(EcalEvent_t * event){

# ifdef PERFORMANCES
  if(firststage == 0){
    ChronoManager CM;
    filling    = CM.createChrono("Cls: Filling");
    init       = CM.createChrono("Cls: Start");
    clustering = CM.createChrono("Cls: Clustering");
    merging    = CM.createChrono("Cls: Merging   ");
    bankwrite  = CM.createChrono("Cls:Bank Wrt");
    firststage = CM.createChrono("All First Stage");
  }
  firststage->start();
# endif
  /////////////
  // start testing if the GCAX table is OK

  if( ArtePointer<GCAX>(ECalInfo.ecalp->fpl)->fcl==0){
    // rebuild the table
    static int wgcax = 0xECA0002;
    arteMsg<<"CAREEVNT1: GCAX table wrong! ";
    arteMsg.warn(wgcax);
    ELLUTsManager::RebuildGCAX();
  }

  //////////////////////////////////////////////////////////////////////////

# ifdef PERFORMANCES
  filling->start();
# endif
  pSignalCells->fill(event);  // Fills the array of pointers

# ifdef PERFORMANCES
  filling->stop();
# endif
  if( CmdCareDump::get(HFLAG) )
    pSignalCells->dump(); // for tests: Dump array 
# ifdef PERFORMANCES
  init->start();
# endif

  //
  // Loop on pSignalCells with the help of the builder
  //
  // initialize loop processing: sort in energy
  //
 
  clusBuilder->startClustering(*pSignalCells, *topClusterTree);

# ifdef PERFORMANCES
  init->stop();
  if( CmdCareDump::get(HFLAG) )
    pSignalCells->dump(); // for tests: Dump array 
  clustering->start();
# endif

  //
  // Clustering LOOP: loop on HITC cells
  //
  unsigned int cells = CmdCareLevel::ncells();
  // test if there are more cells to consider
  while( clusBuilder->more() && cells>0 ){    
    clusBuilder->doClusterize();  // Clusterize with current cell
    clusBuilder->next();       // consider next cell to clusterize
    cells--;
  };

# ifdef PERFORMANCES
  clustering->stop();
  merging->start();
# endif

  //
  // Analyze the Top Cluster List to find the number of hypothetical 
  // particle giving rise to a cluster
  //
  EDClusHypotModifier  clusHypMd;
  clusHypMd.doModify(topClusterTree);

# ifdef PERFORMANCES
  merging->stop();
  bankwrite->start();
# endif
  //
  // Now store the informations on shower hypotheses on the RCCL bank
  //
  EAHypAnalyzer hypAn(showerArray);
  hypAn.doAnalysis(topClusterTree);

  if( CmdCareDump::get(CFLAG) )
    topClusterTree->dump();

# ifdef PERFORMANCES
  bankwrite->stop();
  firststage->stop();
# endif

  return showerArray;
}

//--
/// Find and return all signals detected on ECAL
const vector<const EMCluster*> & ECALDetector::findAllSignals(){

# ifdef PERFORMANCES
  if(secondstage == 0){
    ChronoManager CM;
    secondstage = CM.createChrono("Second Stage");
    matching    = CM.createChrono("Care Match  ");    
    matching1   = CM.createChrono("Care Match1  ");    
    extrap      = CM.createChrono("Marple extrap");
    matching2   = CM.createChrono("Care Match2  ");    
    matching3   = CM.createChrono("Care Match3  ");
    splitting   = CM.createChrono("Splitting");
    bankupdate  = CM.createChrono("Bank update ");
  }
  secondstage->start();
  matching->start();
  matching1->start();
# endif

  //
  // Extrapolate charged tracks to the surface of the calorimeter
  // and find in which cluster the extrapolated point is.
  // Add to the cluster tree infos from tracks
  //

  clusBuilder->trackClusterMatch();
# ifdef WRITE
  arteMsg<<"After Track Cluster Match";
  arteMsg.allp();
# endif
#   ifdef PERFORMANCES
  matching1->stop();
  matching2->start();
# endif
  //
  // Analyze the Top Cluster List to find the number of hypothetical 
  // charged and neutral particles
  // In doing this the number of hypothetical particle can only increase and
  // can never decrease. At the end of the analysis the number of hypotetical
  // particles is given by the sum of charged and neutral clusters
  //

  EDClusTraHypotModifier  clusTraHypMd(clusBuilder);
  clusTraHypMd.doModify(topClusterTree);
  //
  //
# ifdef WRITE
  arteMsg<<"Cluster List Hypotheses: ";
  topClusterTree->getAllHypotheses().dump();
# endif
# ifdef PERFORMANCES
  matching2->stop();
  matching3->start();
# endif
  //
  // An attempt to recover those tracks whose extrapolation point does not
  // belong to any hit cell is done. 
  //
  clusBuilder->recoverUnmatchedTracks();
# ifdef PERFORMANCES
  matching3->stop();
  matching->stop();
# endif
# ifdef WRITE
  arteMsg<<"After Recover Unmatched Tracks";
  arteMsg.allp();
# endif

  //
  // At this point boundary region cells should be recovered
  // 
  if( CmdCareLevel::level()>3){
#   ifdef PERFORMANCES
    splitting->start();
#   endif
    /// split cells only at level 4 and 5
    EDBoundCellModifier  clusBoundCellMd(clusBuilder);
    clusBoundCellMd.doModify(topClusterTree);
#   ifdef PERFORMANCES
    splitting->stop();
#   endif
  }
  //
  //if( CmdCareDump::get(CTFLAG) )
  //    topClusterTree->dump();
  //
  // At this point all clusters are labelled as neutral, single charge
  // or many charges
  //
# ifdef PERFORMANCES
  bankupdate->start();
# endif
  EAHypTraAnalyzer hypTraAn(showerArray, signalArray);
  hypTraAn.analyzeTopList(*topClusterTree);

# ifdef PERFORMANCES
  bankupdate->stop();
  secondstage->stop();
# endif
 
# ifdef WRITE 
  arteMsg<<"Cluster List Hypotheses: ";
  //  topClusterTree->getAllHypotheses().dump();

  if(topClusterTree->getTracks() >= topClusterTree->getChargHyp() &&
     topClusterTree->getHypotheses()==
     topClusterTree->getNeuHyp()+topClusterTree->getChargHyp() ){
    arteMsg<<" GOOD TOP Cluster List";
    arteMsg.allp();
  } else{
    arteMsg<<" BAD TOP Cluster List";
    arteMsg.allp();
    if( !CmdCareDump::get(CTFLAG) )
      topClusterTree->dump();
  }
# endif

  if( CmdCareDump::get(CTFLAG) )
    topClusterTree->dump();
  
  return signalArray;
} // find All Signals


//
// Store neutrals into the RTRA bank
void ECALDetector::storeNeutrals(){
# ifndef ONLINE
  if( ArtePointer<RCCL>::getLink()()!=0){
    for(ArteTableIterator<RCCL> rccl; rccl.valid(); rccl.next()){
      ArtePointer<RCLX> rclx(rccl);
      // Neutral clusters are defined as those without segments connected
      if( !ArteRelationIterator<RCLX_RSEG>(rclx->rseg).valid() ){ // test for connected segments
	// get a table entry
	MMBank<RTRA>::Book(1);
	ArtePointer<RTRA> rtra(ArtePointer<RTRA>::nrow());
	// Now copy all the relevant informations
	rtra->zf = rccl->z;
	rtra->xf = rccl->x;
	rtra->yf = rccl->y;
	rtra->txf = rccl->tx;
	rtra->tyf = rccl->ty;
	rtra->pf = 1./rccl->esum;
	for(int i=0; i<14; i++){
	  rtra->cvf[i] = rccl->cve[i];
	}
	float tmp = rccl->esum*rccl->esum;
	rtra->cvf[14] = rccl->cve[14]/(tmp*tmp);
	rtra->hca = (int) (rccl->wid);
	// Now I've finished with RTRA. Let's begin with RTRX
	// ArtePointer<RTRX> rtrx(rtra);
	unsigned int nrc = rccl.index();
	unsigned int ntr = rtra.index();
	//	cout<<" nTRA: "<<ntr<<"  RCCL: "<<nrc<<endl;
	MMBank<RTRX_RCCL>::Brel( ntr, nrc);
	//	mmbrel_(ArtePointer<RTRX_RCCL>::getLink().getRef(), ntr, nrc);
	for(ArteRelationIterator<RCLX_HITC> hitc(rclx->hitc); hitc.valid();
	    hitc.next()){
	  unsigned int nhc = ((ArtePointer<HITC>)hitc).index();
	  MMBank<RTRX_HITC>::Brel( ntr, nhc);
	}
      }
    }
  }
# endif
}



void ECALDetector::careevnt1() {

  if( !ECalInfo.init){
    //
    //  careevnt1 called with uninitialized reconstruction
    //
    arteMsg<<"Careevnt1() called without prior initialization:"
	   <<" check your code !";
    arteMsg.error();
  }
  instance()->findShowers(0);
}

void ECALDetector::careevnt2() {

  if( !ECalInfo.init){
    //
    //  careevnt2 called with uninitialized reconstruction
    //
    arteMsg<<"Careevnt2_() called without prior initialization:"
	   <<" check your code !";
    arteMsg.error();
  }
  instance()->findAllSignals();
}

void ECALDetector::stop() {
 
  if( ECalInfo.init ){
    //
    //
#   ifdef PERFORMANCES
      ChronoManager CM;
      CM.printInformations();
#   endif
    arteMsg<<stars;
    arteMsg.allp();
    arteMsg<<" Stop of Care  "<<VERSION;
    arteMsg.allp();
    arteMsg<<stars;
    arteMsg.allp();
    // set to false the initialization flag
    ECalInfo.init = false;
  } else {
    //
    //  carestop called when the calorimeter was not initialized
    //
    arteMsg<<"STOP for CARE called without a previous initialization call";
    arteMsg.allp();
  }
}

//--
/// Store parameters of a track
void  ECALDetector::storeTrack(TrackInfo & trackInfo){
  instance()->topClusterTree->newTrack(trackInfo);
}
