//                              -*- Mode: C++ -*-
//: ESPID04.C
//
//
// Author: roberto spighi   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: spighi@bo.infn.it
//
// Created 4/7/97
// Last Update ........
//

#include "care/EMCluster.hh"
#include "care/EMHitCell.hh"
#include "care/ESPID04.hh"
#include <math.h>

/* 
CLASS
  ESPID04
   Class for PID (pion/electron discrimination) analysis: 
*/

// parameter of likelihood initializzation

float ESPID04::eparinner[9]={8.590, 1.4967, 0.1375, 0.63451, 1.5409, 0.94375,
			     0.19489, 16.131, -7.3106};
float ESPID04::pparinner[9]={1.69686, 1.51615, 0.0645098, 0.110667, 
			     1.87817,1.31826, 0.70271, 12.8721, -9.03208};
float ESPID04::eparmiddle[9]={-16.165, 2.92583, 0.474919, 0.934367, 40.3047,
			      3.82188, 1.05724, 13.7673, -8.49681};
float ESPID04::pparmiddle[9]={24.795, 3.5825, .63092, .049049, 327.73,  
			      5.1922, 1.8252, .62884, -.70174};
float ESPID04::eparouter[9]={169.56, 6.6594, 1.0606, 2.3821, 178.583,
			     5.5864, 1.64282, 5.28654, -5.15417};
float ESPID04::pparouter[9]={48.94, 9.46202, 3.03828, 4.80237, 17.7482,
			     5.45011, 1.16202,  6.95742, 31.4483};

ESPID04::ESPID04() : ESPIDStrategy() {}

ESPID04::~ESPID04(){}

 
//--
//  Start a cluster analysis
void ESPID04::beginAnalysis(const EMCluster* clus, ArtePointer<RCCL> prccl,
			    const EMHitCell* phitc){
  
  ESPIDStrategy::beginAnalysis(clus, prccl, phitc);
  energySum04 = 0.;
  rsum04 = 0.;

  doAnalysis(clus);

  if(energySum04>0.) {
    RPID = rsum04/energySum04;
  }
  else{
    RPID = -1.;
  }
  // Evaluation likelihoods for electron and hadron
  likeElectron();

  likeHadron();
  updatebank(rccl);

}

//--
// HitCell Analyzer
void ESPID04::doAnalysis(const EMHitCell* tower){

  const float power = 0.4;

  // for a tower add it's energy
  Energy_t  ener = tower->energy();
  Energy_t ener04;
  
  float distcogcell=sqrt((tower->x() - xcog)*(tower->x() - xcog) + 
			 (tower->y() - ycog)*(tower->y() - ycog));

  ener04 = pow(ener, power);
  energySum04 += ener04;
  rsum04 +=distcogcell*ener04;
}


//--
// Cluster Analyzer
void ESPID04::doAnalysis(const EMCluster * cluster){

  // loop on clusters
  
  for(EMClusterMember* child = cluster->begin(); 
      child!=0;  child = child->next() ){
    child->analyzeMe(*this);
  }

}
void ESPID04::likeElectron(){
  //  function to calculate the likelihood for electron hypotesis
  float biexp;
  float gau;
  float x;
  float espongau;
  float epar[9];

  if(typecal==5) 
    for(int i=0; i<9; i++)
      epar[i] = eparinner[i];

  if(typecal==2)
    for(int i=0; i<9; i++)
       epar[i] = eparmiddle[i];
  if(typecal==1) 
    for(int i=0; i<9; i++)
       epar[i] = eparouter[i];

  x=RPID;
  if( x>0.1){
    float argpow = fabs((x-epar[1])/epar[2]);
    float power = -pow(argpow,epar[3]);
    if(abs(power)<70.0){
      biexp = epar[0]*exp(power);
    } else {
      biexp = 0.0;
    }
    espongau = (x-epar[5])/(sqrt(2.0)*epar[6]);
  
    if( espongau<8.0 ){
      gau =  epar[4]*exp(-espongau*espongau);
    } else {
      gau = 0.0;
    }
    likeEle = (biexp + gau)*(epar[7]/x + epar[8]/(x*x));
    if(likeEle<0.) likeEle = 0.;
  } else {
    if( typecal==1 ) likeEle = 360.;
    if( typecal==2 ) likeEle = 75.;
    if( typecal==5 ) likeEle = 17.;
  }
}



void ESPID04::likeHadron(){
  //  function to calculate the likelihood for electron hypotesis
  float biexp;
  float gau;
  float x;
  float espongau;
  float ppar[9];

  if(typecal==5) 
    for(int i=0; i<9; i++)
      ppar[i] = pparinner[i];

  if(typecal==2)
    for(int i=0; i<9; i++)
      ppar[i] = pparmiddle[i];

  if(typecal==1) 
    for(int i=0; i<9; i++)
      ppar[i] = pparouter[i];

  x=RPID;

  if( x>0.1){
    float argpow = fabs((x-ppar[1])/ppar[2]);
    float power = -pow(argpow,ppar[3]);
    if(abs(power)<70.0){
      biexp = ppar[0]*exp(power);
    } else {
      biexp = 0.0;
    }
    espongau = (x-ppar[5])/(sqrt(2.0)*ppar[6]);
    if( espongau<8.0 ){
      gau =  ppar[4]*exp(-espongau*espongau);
    } else {
      gau = 0.0;
    }
    likeHad = (biexp + gau)*(ppar[7]/x + ppar[8]/(x*x));
    if(likeHad<0.) likeHad = 0.;
  } else {
    if(typecal==1 ) likeHad = 3700.;
    if(typecal==2 ) likeHad = 1250.;
    if(typecal==5 ) likeHad = 140.;
  }
}


void ESPID04::setparlike04(int cal,int particle,float like04[]){
  //  set the parameter for likelihood curves of electron and pion 
  if(cal==1 && particle==1) 
    for(int i=0; i<9; i++ ) 
      eparinner[i] = like04[i];

  if(cal==2 && particle==1) 
    for(int i=0; i<9; i++ ) 
      eparmiddle[i] = like04[i];

  if(cal==3 && particle==1) 
    for(int i=0; i<9; i++ ) 
      eparouter[i] = like04[i];


  if(cal==1 && particle==2) 
    for(int i=0; i<9; i++ ) 
      pparinner[i] = like04[i];

  if(cal==2 && particle==2) 
    for(int i=0; i<9; i++ ) 
      pparmiddle[i] = like04[i];

  if(cal==3 && particle==2) 
    for(int i=0; i<9; i++ ) 
      pparouter[i] = like04[i];

}

//--
// ClusterTree Analyzer
void ESPID04::doAnalysis(const EMClusterTree *){}

//--
// Cluster Analyzer
void ESPID04::doAnalysis(const EMClusterMember *){}
//--
// Track Analyzer
void ESPID04::doAnalysis(const EMTrackOnEcal *){}









































