#ifndef DATA_STREAM_HH
#define DATA_STREAM_HH

#include <sys/types.h>
#include <netinet/in.h>
#include "gpack/gpack.h"
#include "gpack/dataforms.h"
#include "gpack/shmstream.hh"
#include "gpack/filestream.hh"

/*   Event is a container object containing other objects: Tables, TableDefs.
     Event starts with EventHeader and finishes with EventTail.
       
         The standard layout of the event produced by L4Farm node and
	 shipped to the Data Logger is the following:

      +---   Event Header
      |      Table Definition 1      
      |      Table 1
      |      ....
      |      Table Definition N
      |      Table N
      |      Event Tail
      +---
        The layout of the event stored on disk:

	     Event Header
             Table 1
             ....
	     Table N
             Event Tail

   The Table Definitions are written once per file.  Corresponding
   Table headers have references to the Table Defs at the beginning of the
   file.

   All objects have prefixes: length of object and ID identifying the
   type of object.

   Table Definition Entry  looks like:

    length   - multiple to sizeof(int)     
    type     - identifies this object as a table definition
    namelen
    name       table name including null ('\0') character
    desc       description = list of field_name:format separated
                        by space.


   Table looks like:

   length   - multiple to sizeof(int)
   type       identifies this object as a table
   pos      - position of the related Table Def in the stream
   ncols    - number of columns  
   nrows    - number of rows
*/

typedef enum WriteStatus 
{ 
  HDR_OK  = 0x01, 
  DEF_OK  = 0x02,
  TAB_OK  = 0x04,
  DATA_OK = 0x08,
  TAIL_OK = 0x10
} WriteStatus;

template <class T> 
class datastream
{

private:

  T           *stream;
  StreamType  streamtype;
  int         flags;        // Control of function call chain
  size_t      eventPos;     // Current Event position on the stream
  int         eventSize;    // Size of current event
  int         eventLength;  // # of bytes of current event not read yet in 
  EventHdr    hdr;
  Origin      local;
  OriginFP    localFP;
  Origin      target;
  OriginFP    targetFP;
  TableDef    *globalList;
  Table       tab;
  int         tablePos;     // position of current table header
  int         tableSize;    // table size
  TableDef    *def;         // pointer to TableDef for current table
  int         dataLength;   // # of bytes of current table not read yet
  EventTail   tail;
  unsigned    *tmpbuf;       // Auxilliary buffer for conversions
  unsigned    arraySize;     // The size of that buffer
  char        errmsg[256];

  void        SwapArray(unsigned *array, int narray);
  void        ConvertIEEE2GFloat(unsigned *array, int narray);
  void        ConvertGFloat2IEEE(unsigned *array, int narray);
  int         IntEncode(Item *it, int *in, int *out, int n, int off);
  int         IntDecode(int *in, int *out, int n, int off);
  int         FloatEncode(Item *it,
			  unsigned *input,
			  unsigned *output,
			  int n,
			  int offset);
  int         FloatDecode(unsigned *input,
			  unsigned *output,
			  int n,
			  int offset);
  int         HollerithEncode(Item *it,
			      unsigned *input,
			      unsigned *output,
			      int n,
			      int offset);
  int         HollerithDecode(Item *it,
			      unsigned *input,
			      unsigned *output,
			      int n,
			      int offset);
  int         EncodeItem(Item* it, 
			 unsigned* input,
			 unsigned* output,
			 int nrows,
			 int rowSize);
  int         DecodeItem(Item *it,
			 unsigned *input,
			 unsigned *output,
			 int nrows,
			 int rowSize);
  void        AddToList(TableDef *def);
  TableDef*   FindPositionInList(size_t pos);
  TableDef*   FindNameInList(const char* name);
  int         NumberOfElements(char *formats);
  int         SetItem(char* source, Item *it);
  int         SetField(char* source, Field *f);
  int         TranslateDesc(char* source, Formats *f);
  int         CheckTableDef();

public:

              datastream();
             ~datastream();
  char*       getErrorMsg();
  int         settarget(TargetType t);
  void        setbuf(int size);
  int         setname(char* name, int mode);
  char*       getname() { return stream->getname(); }
  void        setkey(int key);
  int         create();
  int         destroy();
  int         open();
  int         close();
  int         is_open() { return stream->is_open(); }
  int         setpos(size_t pos) { return stream->setpos(pos); }
  StreamType  getStreamType();
  void        setevtype(EventType type);
  EventType   getevtype(); 
  int         getevtpos();
  int         getevt(EventHdr& eh);
  int         putevt(EventHdr& eh);
  int         gettabinfo(char *name, int& ncols, int& nrows);
  int         skiptab();
  int         gettab(char* name,
		     int   ncols,
		     int   nrows,
		     char* desc,
		     void* data);
  int         puttab(const char* name,
		     int ncols,
		     int nrows, 
		     const char* desc,
		     void* data);
  int         flushevt();
};

typedef datastream<filestream>  File_Stream;
typedef datastream<shmstream>   Shared_Memory;

typedef union 
{
      float f;
      unsigned int u;
} FC;

#include "gpack/datastream.icc"

#endif




