//**********************************************************************/
// Created: 23.08.97,   Olya Igonkina   for   MARPLE package
//
// implementation of RcMatchPoint class
//
//**********************************************************************/

#include "marple/marple.hh"
#include <CLHEP/Matrix/Vector.h>
#include "auto/GEDE.hh"

#include "arte/ArteMessage.hh"

bool comp_chi2( const RcMatchPoint * prm1, const RcMatchPoint * prm2)
{  return prm1->chi2() < prm2->chi2(); }

RcMatchPoint::RcMatchPoint( RcPoint &rp1, 
			    RcPoint &rp2,
			    int lineswim = 0)
  : RcPoint(), foriginPoints(), fHitbList()
{
  RcPoint* ptPoint=0;
  int createTmp=0;
  if( abs(rp1.z() - rp2.z()) > 0.1e-4 )
    {
      // in case rp1.z() != rp2.z(), prolongate rp1 to rp2.z()
      ptPoint = new RcPoint(rp1);
      if( lineswim )
	{
	  ptPoint->lineswim(  rp2.z() );
	}else{
	  ptPoint->prolongate( rp2.z() );
	}
      createTmp = 1;
    }else{
      ptPoint = &rp1;
    }
  setZ( rp2.z());  
  setmark(0);
  setchi2(MARPLE_INFINITY);

  //find all HITB's
  fHitbList =  rp1.hitbList() ;
  unsigned int oldSize = hitbList().size();

  // supposed that every entry in hitbList must appear only one time in list  
  vector<ArtePointer<HITB> > addHitb = rp2.hitbList();
  for( unsigned int i=0; i < addHitb.size(); i++ )
    {
      int newHitb = 1;
      for(unsigned int j=0; j < oldSize && newHitb ; j++ )
	{ 
	  if( addHitb[i] == hitbList()[j] ){ newHitb = 0 ;}
	}
      if( newHitb ){ fHitbList.push_back(addHitb[i]); }
    }
  
  int ifail;
  HepVector p1(5);
  p1(1) = (*ptPoint).x();
  p1(2) = (*ptPoint).y();
  p1(3) = (*ptPoint).tx();
  p1(4) = (*ptPoint).ty();
  p1(5) = (*ptPoint).pe();
  HepSymMatrix invCov_1 = (*ptPoint).cov().inverse(ifail);
  if( ifail != 0 ){ error_message();   setmark(ifail); return;}

  HepVector p2(5);
  p2(1) = rp2.x();
  p2(2) = rp2.y();
  p2(3) = rp2.tx();
  p2(4) = rp2.ty();
  p2(5) = rp2.pe();
  HepSymMatrix invCov_2 = rp2.cov().inverse(ifail);
  if( ifail != 0 ){ error_message();  setmark(ifail); return; }


  // calculate the matrix 1/Cov = 1/Cov1 + 1/Cov2 

  HepSymMatrix cov_match = invCov_1 + invCov_2;
  cov_match.invert(ifail);
  if( ifail != 0 ){ error_message(); setmark(ifail); return; }

  setcov( cov_match);

  // calculate the mean parameters

  HepVector pmean = invCov_1 * p1  ;
  pmean += (invCov_2*p2);
  pmean = cov_match * pmean; 

  setX(  pmean(1));
  setY(  pmean(2));
  settx( pmean(3));
  setty( pmean(4));
  setpe( pmean(5));

  // calculate the chi2 of matching
  p1 -= pmean;
  p2 -= pmean;
  setchi2( invCov_1.similarity( p1) +  invCov_2.similarity( p2)); 
  setquality( 0 );

  // save origin
  //RcPoint* address = &rp1;
  foriginPoints.push_back( &rp1 );
  //address = &rp2;
  foriginPoints.push_back( &rp2 );

  if( createTmp ) 
    {
      delete ptPoint;
    }
  ptPoint = 0;
}
//----------------------------------------------------------------
//  correct RcMatchPoint by measured HITB
void RcMatchPoint::addHitb( const ArtePointer<HITB> hitb, int & iErr )
{
  iErr = 0;
  int newHit = 1;
  for( unsigned int i=0; i < fHitbList.size(); i++ )
    {
      if( fHitbList[i] == hitb ){ newHit = 0; }
    }
  if( !newHit )
    {
      arteMsg << "MARPLE: HITB " << hitb 
	   << " already connected with that RcMatchPoint";
      arteMsg.allp();
      return;
    }
  
  fHitbList.push_back(hitb);
  HepDouble deltaChi2 = RcPoint::filter( hitb, iErr);
  if( !iErr )
    {
      setchi2( chi2() + deltaChi2 );
      // svd_f_chi2 is taken from Joerg note = 0.2
      HepDouble svd_f_chi2 = 0.2;
      setquality( quality() + 1 - svd_f_chi2 * deltaChi2 );  
    }else{
      arteMsg << "MARPLE: addHitb: unable to filter RcMatchPoint to HITB" 
	      << hitb ;
      arteMsg.allp();
    }
}
//----------------------------------------------------------------

//  Kalman refit of the one of original RcPoint's using RcMatchPoint parameters
HepDouble RcMatchPoint::refitOrig(int & iErr, int iOrig )
{
  if( iOrig >= origin().size() ) 
    {
      arteMsg << "MARPLE: refitOrig : only " << origin().size() 
	      << " RcPoint connected with RcMatchPoint object" ;
      arteMsg.allp();
      return 0;
    }
  HepDouble deltaChi2 = 0;
  iErr = 0;
  int iErr1 = 0;
  if( iOrig >= 0 )
    {
      deltaChi2 += origin()[iOrig]->refit(iErr1);
      if( iErr1 ) iErr=iErr1;
    }else{
      for( unsigned int iCount=0; iCount < origin().size(); iCount++ )
	{
	  deltaChi2 += origin()[iCount]->refit(iErr1);
	  if( iErr1 ) iErr=iErr1;
	}
    }
  return deltaChi2;
}

//----------------------------------------------------------------

// returns relation with MTRA 
vector<ArtePointer<MTRA> > RcMatchPoint::mtraList() const
{
  vector<ArtePointer<MTRA> > mList;
  for( unsigned int i=0; i < origin().size(); i++ )
    {
      vector<ArtePointer<MTRA> > nextList( (*origin()[i]).mtraList() );
      for( unsigned int j=0; j < nextList.size(); j++ )
  	{
	  mList.push_back(nextList[j]);
	}
   }
  return mList;
}

//----------------------------------------------------------------

// returns relation with HITC 
vector<ArtePointer<HITC> > RcMatchPoint::hitcList() const
{
  vector<ArtePointer<HITC> > hList;
  for( unsigned int i=0; i < origin().size(); i++ )
    {
      vector<ArtePointer<HITC> > nextList( (*origin()[i]).hitcList() );
      for( unsigned int j=0; j < nextList.size(); j++ )
  	{
	  hList.push_back(nextList[j]);
	}
   }
  return hList;
}

ostream & operator << (ostream & os, const RcMatchPoint &rp)
{
  os << "Match Point with  chi2 = " << rp.chi2() 
     << " , Quality = " << rp.quality() <<  endl;
  const RcPoint * ptPoint = &rp;
  os << *ptPoint;
  return os ;
}
