#ifndef _RCMATCHBUILDER_HH_
#define _RCMATCHBUILDER_HH_
//---------------------------------------------------------------------
// Created: 13.10.97,   Olya Igonkina

// MARPLE builder class of Points for Reconstruction

// !!! important
// BE SURE that the classes used for matching are derived from RcPoint class
// parameters of match will not deleted

//---------------------------------------------------------------------


#include <vector>
#include <iterator>

#include "marple/RcPoint.hh"
#include "marple/RcMatchPoint.hh"
#include "marple/RcMatchConst.h"

/** {\bf MARPLE main class to do match }                        \\
    The procedure to do match is described in section \Ref{Usage} of
    this manual.

    In addition the following functions present the result of matching:     \\
    vectorLeft().size()  - number of RcPoints to be match from left side    \\
    vectorRight().size() - number of RcPoints to be match from right side   \\
    NrealPair()          - number of pairs of RcPoints saved in left and right
                           vectors corresponded to one MTRA particle,       \\
    matchList().size()   - number of matched pairs                          \\
    Nbad()               - number of matched pairs of RcPoints corresponded 
                           to the different MTRA particles                  \\
    Nlost()              - number of lost during match pairs of RcPoints    
                           correspond to the one MTRA particle              \\

   Two {\bf important} points should be keeped in mind:
   \begin{itemize}
   \item The class RcMatchBuilder is a template one. It means that 
   to use the object of RcMatchBuilder type the two classes  
   {\bf PointsLeft} and {\bf PointsRight} should be defined.
   For correct usage of RcMatchBuilder they should be inherited 
   from the \Ref{RcPoint} class. The {\it mark} member field of RcPoint class 
   is used to manage the number of combination with the given object.
   \item To avoid garbage creation the cleaning functions should be 
   called in time. The best way is to invoke {\bf cleanAll()} function when 
   the object is not necessary anymore. Do it always at the end of the
   logical slope of program. 
   \end{itemize}

   @memo MARPLE main class to do match 
*/
template <class PointsLeft, class PointsRight>
class RcMatchBuilder
{
public:
  ///              default construtor
  inline RcMatchBuilder(); 

  ///              make match of two vectors with rcConst constants
  RcMatchBuilder( vector<PointsLeft *> vecLeft,
		  vector<PointsRight *> vecRight,
		  RcMatchConst & rcConst);

  ///               destructor
  inline ~RcMatchBuilder();

  ///               put left vector of RcPoints to be match
  inline void addLeft(  vector<PointsLeft *> & vecLeft);
  ///               put right vector of RcPoints to be match
  inline void addRight( vector<PointsRight *> & vecRight);

  ///               add Rcpoint to left vector to be match
  inline void addPointLeft(  PointsLeft & pointLeft);
  ///               add Rcpoint to right vector to be match
  inline void addPointRight( PointsRight & pointRight);

  ///               set constants for matching see \Ref{RcMatchConst}
  inline void setConst( RcMatchConst & rcConst)
    { fptRcConst = &rcConst; }; 

  ///               returns left vector to be match
  inline vector< PointsLeft  *> vectorLeft()  const
    { return fvecLeft;   };
  ///               returns right vector to be match
  inline vector< PointsRight *> vectorRight() const
    { return fvecRight;  };

  ///               returns left point connected with MatchPoint iMatch in the matchList,  0 <= iMatch < matchList().size() 
  inline PointsLeft * leftOrigin( unsigned int iMatch ) const;

  ///               returns right point connected with MatchPoint iMatch in the matchList,  0 <= iMatch < matchList().size() 
  inline PointsRight * rightOrigin( unsigned int iMatch ) const;
  
  ///       returns vector of RcMatchPoints with result of matching to be match
  inline vector< RcMatchPoint *> matchList()   const
    { return fmatchList; };
  ///               returns constants for matching
  inline RcMatchConst  * ptToMatchConst() const { return fptRcConst; };

  ///              returns real number of pairs correspond to one MTRA particle
  inline int    NrealPair()      const{ return fNrealPair; };
  ///               returns number of wrong pair
  inline int    Nbad()           const{ return fNbad;      }; 
  ///               returns the number of lost pair correspond to one MTRA particle
  inline int    Nlost()          const{ return fNlost;     }; 

  //main operation :

  ///               make match between vectorLeft and vectorRight
  void doMatch();  
  ///               print information about matching
  void stat(int iLong = 1)   const; 
  ///               fill ntuple Deltas with cuts
  void fillDeltas(int iNtup = 0) const; 
  ///               fill ntuple with statistics
  void fillMatch( int iNtup = 0) const; 

  ///Kalman refit of left matched objects using mean parameters from RcMatchPoint
  void refitLeftFound(int & iErr);

  ///Kalman refit of left matched objects using mean parameters from RcMatchPoint
  void refitRightFound(int & iErr);


  void updateARTELeftFound(int & iErr) const;
  void updateARTERightFound(int & iErr) const;

  void saveRTRA() const;            // save information into ARTE banks
  void saveRSEG() const;
  void saveRCCL() const;
  /// save relation between RcPoints from left and right vectors which pass match procedure, return 0 in case of save errors 
  void saveRelation(int & iErr) const;
  void saveRelation(const PointsLeft * orig1, 
		    const PointsRight * orig2, 
		    int & iErr) const;

  //delete functions:  iStat = 1 delete statistics 

  //inline void deleteLeftPoint(  int iStat = 1);  // delete information
  //inline void deleteRightPoint(  int iStat = 1);  // delete information
  ///                 delete information saved in left vector
  inline void deleteLeft(  int iStat = 1);  // delete information
  ///                 delete information saved in Right vector
  inline void deleteRight( int iStat = 1);  
  ///                 delete match information 
  inline void deleteMatch( int iStat = 1);  
  /// delete all information if iStat = 1 statistics of match is also cleaned
  inline void deleteAll(   int iStat = 1);
  /// delete all information if iStat = 1 statistics of match is also cleaned
  inline void clearAll(   int iStat = 1);
  //                  clean statistics information
  inline void deleteStat();

  ///                 copy all information saved in RcMatchBuilder object
  inline RcMatchBuilder< PointsLeft, PointsRight> & operator = 
  (const RcMatchBuilder< PointsLeft, PointsRight> &rp );

protected:
  ///                     left vector to be match
  vector< PointsLeft   * >  fvecLeft;
  ///                     right vector to be match
  vector< PointsRight  * >  fvecRight;
  ///                     vector with match information, filled by doMatch()
  vector< RcMatchPoint * >  fmatchList;
  ///                     constants used by matching
  RcMatchConst *            fptRcConst;

  ///   number of pairs of RcPoints from left and right vectors correspond to one MTRA particle
  int fNrealPair;
  ///           number of ghost pairs found by matching    
  int fNbad;
  ///           number of lost pairs during matching
  int fNlost;
};

//template <class PointsLeft, class PointsRight>
//ostream & operator << (ostream &, const RcMatchBuilder< PointsLeft, 
//		                                        PointsRight> &);

#include "marple/RcMatchBuilder.icc"

//unfortunately has to include that also to solve problem with template 
#if __GNUC__ <= 2 && __GNUC_MINOR__ <= 7
#include "marple/RcMatchBuilder.C"
#endif

#endif
