// ===========================================================================
//
//   SingleSphericalMirror
//

#include "rich/vessel.h"
#include "rich/Losses.h"

extern int RICHdebug;

// -------------------------------------------------------------------------
//
//  get intersection point
//

int SingleSphericalMirror::intersection(const StraightLine& p,double *path) {
  Vec3   v =p.x()-coc;
  double v2=v*v;
  if (v2>radius()*radius()) return SMI_LOSS_OUTOFSPHERE; // out of sphere
  double vn=v*p.n();
  *path=-vn+sqrt(radius()*radius()-v2+vn*vn);
  return NO_LOSS;
  }

// -------------------------------------------------------------------------
//
//  reflexion & reflexion loss
//

int SingleSphericalMirror::reflect(Photon* p) {
  int ret;
  double path;
  if ((ret=intersection(*p,&path))) return ret;
  if ((ret=RICH_Vessel->trace(p,path))) return ret;

//
//  throw photon away, if mirror is not hit
//
  if (!inside(TransformFromOld(p->x()))) return SMI_LOSS_MIRRORGAPS;

// reflexion

  Vec3 ns=normalize(p->x()-coc);      // normal vector at hit position
  p->n()-=double(2)*(ns*p->n())*ns;  // short hand formula for the reflection
//
//  treat reflexion loss
//
  if (rnd_normal()>reflectivity().f(p->lambda())) return SMI_LOSS_REFLECTIVITY;
  return NO_LOSS;
  }

// -------------------------------------------------------------------------
//
//   reflexion used to find ring center
//


int SingleSphericalMirror::reflect(StraightLine* t) {
  int ret;
  double path;
  if ((ret=intersection(*t,&path))) return ret;
  t->x()+=path*t->n();
// reflexion

  Vec3 ns=normalize(t->x()-coc);     // normal vector at hit position
  t->n()-=double(2)*(ns*t->n())*ns;  // short hand formula for the reflection

  return NO_LOSS;
  }


// -------------------------------------------------------------------------
//
//  check if distance to plane has always the same sign <=> point is inside 
//  v has to be given in local coordinates
//

int SingleSphericalMirror::inside(const Vec3& v) {
  if (numcorner==0) return 1;  
  if (fabs(v.x())>xdim      ) return 0;
  if (fabs(v.y())>halfdiameter()) return 0;
  int i;
  int is=1,side=0; 
  for (i=0;i<numcorner;i++) {
    float s = nbound[i]*v+dbound[i];
    is= (s<0) ? -1 : 1;
    if (side==0) side=is;
    else if (!(is==side)) break;
    }
  return (is==side);
  }


// -------------------------------------------------------------------------
//
//    initialisation from ARTE Pointer  
//

SingleSphericalMirror::SingleSphericalMirror(ArtePointer<CSSM> aa) {
  int i;

  a=aa;
  if (RICHdebug) { printf("init SingleSphericalMirror\n");fflush(stdout);}
  r=a->row;
  c=a->col;

  if (RICHdebug) { printf("index %d %d %d %d\n",r,c,aa->row,aa->col);fflush(stdout);}
  x0()=Vec3(a->orig[0],a->orig[1],a->orig[2]);
  ez()=Vec3(a->norm[0],a->norm[1],a->norm[2]);
  rr=norm(ez());
  ez()/=radius();
  
  ex()=Vec3(a->xvec[0],a->xvec[1],a->xvec[2]);
  xdim=norm(ex());
  ex()/=xdim;

  ey()=cross(ez(),ex());
  coc=x0()+radius()*ez();
    
  dd=a->hdia;
    
  refl=Table(a->iref,a->nref);
  
  memcpy(ttype,a->type,sizeof(a->type));
  for (i=sizeof(a->type)-1;i>=0;i--) if (ttype[i]!=' ') break;
  ttype[i+1]='\0';
  
  memcpy(nname,a->name,sizeof(a->name));
  for (i=sizeof(a->name)-1;i>=0;i--) if (nname[i]!=' ') break;
  nname[i+1]='\0';
  
  numcorner=a->ncor;
  for (i=0;i<numcorner;i++) {
    int j;
    for (j=0;j<3;j++) corner[i][j]=a->corn[i*3+j];
    }
    
  for (i=0;i<numcorner;i++) corner[i]=TransformFromOld(corner[i]);

  for (i=0;i<numcorner;i++) {
    int j= (i==0) ? numcorner-1 : i-1;
    nbound[i][0]= (corner[i][1]-corner[j][1]);
    nbound[i][1]=-(corner[i][0]-corner[j][0]);
    nbound[i][2]= 0;
    dbound[i]=-(corner[i]*nbound[i]);
    }
  if (RICHdebug) {printf("init\n");fflush(stdout);}
  }

// -------------------------------------------------------------------------
//
//    initialisation from input file  
//

SingleSphericalMirror::SingleSphericalMirror(ScriptFile& f,Table& rrefl) :
  a(ArtePointer<CSSM>(ArtePointer<CSSM>::nrow()+1)) {
  char n[256],tag[256];
  int i;

  refl=rrefl;
   
  strcpy(ttype,"\0");
  while (f.getline(n)) {
    char *s=n;
    sscanf(s,"%s",tag);s+=f.itemlength(s,1);
    if (RICHdebug) {printf("%sXXX %s",tag,n);fflush(stdout);}
    int err=0;
    if      (!strcmp(tag,"index")) { 
      err=(sscanf(s,"%d %d",&r,&c)!=2);
      }  
    else if (!strcmp(tag,"type")) {
      err=(sscanf(s,"%s",&ttype[0])!=1);  
      if (strncmp(ttype,"HEX",3)==0)  numcorner=6;
      else if (strncmp(ttype,"FIVE",4)==0) numcorner=5;
      else if (strncmp(ttype,"HALF",4)==0) numcorner=4;
      else if (strncmp(ttype,"FOUR",4)==0) numcorner=4;
      else {
        printf("Unknown mirror type ..: %s\n",ttype);
        exit(1);
        }
      }
    else if (!strcmp(tag,"mirrorname"))   
      err=(sscanf(s,"%s",&nname[0])!=1);      
    else if (!strcmp(tag,"center")) 
      err=(sscanf(s,"%lf %lf %lf",&x0().x(),&x0().y(),&x0().z())!=3);
    else if (!strcmp(tag,"normal")) 
      err=(sscanf(s,"%lf %lf %lf",&ez().x(),&ez().y(),&ez().z())!=3);
    else if (!strcmp(tag,"horizontal")) 
      err=(sscanf(s,"%lf %lf %lf",&ex().x(),&ex().y(),&ex().z())!=3);
    else if (!strcmp(tag,"vertices")) {
      int i;
      for (i=0;i<numcorner;i++) {
        f.getline(n);
        Vec3& c=corner[i];
        err=(sscanf(n,"%lf %lf %lf",&c.x(),&c.y(),&c.z())!=3);
        if (err) break;
        }
      }
    else if (!strcmp(tag,"radiusofcurvature")) 
      err=(sscanf(s,"%lf",&rr)!=1);
//    else if (!strcmp(tag,"horizontal_dimension")) 
//      err=(sscanf(s,"%lf",&xdim)!=1);
    else if (!strcmp(tag,"halfdiameter")) 
      err=(sscanf(s,"%lf",&dd)!=1);
    else if (!strcmp(tag,"end"))  break;
    else {
      fprintf(stderr,
         "SingleSphericalMirror(ScriptFile):  Unknown tag: %s\n",tag);
      exit(0);
      }
/*
    if (err) {
      fprintf(stderr,
         "SingleSphericalMirror(ScriptFile):  Error during scan: %d\n",err);
      exit(0);
      }
*/
    }   
  if (RICHdebug) {print();fflush(stdout);}

  xdim=sqrt(3./4.)*dd;
  CSSM ssm;
  ssm.row=row();
  ssm.col=col();
  ssm.orig[0]=x0().x();
  ssm.orig[1]=x0().y();
  ssm.orig[2]=x0().z();
  ssm.norm[0]=radius()*ez().x();
  ssm.norm[1]=radius()*ez().y();
  ssm.norm[2]=radius()*ez().z();
  ssm.xvec[0]=xdim*ex().x();
  ssm.xvec[1]=xdim*ex().y();
  ssm.xvec[2]=xdim*ex().z();
  ssm.hdia=halfdiameter();
  ssm.iref=reflectivity().firstpoint();
  ssm.nref=reflectivity().numpoints();
  
  ssm.ncor=numcorner;
  for (i=0;i<numcorner;i++) {
    int j;
    for (j=0;j<3;j++) ssm.corn[i*3+j]=corner[i][j];
    }
    
  memset(&ssm.type[0],' ',sizeof(ssm.type));
  int len=(strlen(ttype)>sizeof(ssm.type)) ? sizeof(ssm.type) :strlen(ttype);
  memcpy(&ssm.type[0],ttype,len);

  memset(&ssm.name[0],' ',sizeof(ssm.name));
  len=(strlen(nname)>sizeof(ssm.name)) ? sizeof(ssm.name) :strlen(nname);
  memcpy(&ssm.name[0],nname,len);

  ArtePointer<CSSM>::append(ssm);

  
  if (RICHdebug) {printf("call init\n");fflush(stdout);}
  ey()=cross(ez(),ex());
  coc=x0()+radius()*ez();  

  for (i=0;i<numcorner;i++) corner[i]=TransformFromOld(corner[i]);

  for (i=0;i<numcorner;i++) {
    int j= (i==0) ? numcorner-1 : i-1;
    nbound[i][0]= (corner[i][1]-corner[j][1]);
    nbound[i][1]=-(corner[i][0]-corner[j][0]);
    nbound[i][2]= 0;
    dbound[i]=-(corner[i]*nbound[i]);
    }
  if (RICHdebug) {printf("init\n");fflush(stdout);}
  }


//
//    print data  
//

void SingleSphericalMirror::print(FILE *fp) {
  fprintf(fp,"#\nSingleSphericalMirror\n");
  fprintf(fp,"  index                 %d %d\n",row(),col());
  fprintf(fp,"  type                  %s\n",type());
  fprintf(fp,"  mirrorname            %s\n",name());
  fprintf(fp,"  center                %12f %12f %12f\n",x0().x(),x0().y(),x0().z());
  fprintf(fp,"  normal                %12f %12f %12f\n",ez().x(),ez().y(),ez().z());
  fprintf(fp,"  horizontal            %12f %12f %12f\n",ex().x(),ex().y(),ex().z());
  int i;
  fprintf(fp,"  vertices\n");
  for (i=0;i<numcorner;i++) {
    TransformToOld(corner[i]).print("   %12f %12f %12f\n",fp);
    }
  fprintf(fp,"  radiusofcurvature     %12f\n",radius()); 
//  fprintf(fp,"  horizontal_dimension  %12f\n",xdim); 
  fprintf(fp,"  halfdiameter          %12f\n",halfdiameter()); 
  fprintf(fp,"  end\n"); 
  fflush(fp);
  }


//
//    to have a look at the mirror with paw  
//

void SingleSphericalMirror::xyview(FILE* fp) {
  int i,j=numcorner-1;
  fprintf(fp,"* numcorner %d\n",numcorner);
  for (i=0;i<numcorner;i++) {
    Vec3 c1=TransformToOld(corner[j]);
    Vec3 c2=TransformToOld(corner[i]);
    fprintf(fp,"line %f %f %f %f\n",c1.x(),c1.y(),c2.x(),c2.y());
    j=i;
    }
  }

