#ifndef _PHOTONDETECTION_H__
#define _PHOTONDETECTION_H__

#include "rich/ricutil.h"
#include "rich/Losses.h"
#include "rich/PMType.h"
#include "rich/LensSystem.h"
#include "rich/PM.h"
#include "rich/PhotonDetectorCell.h"
#include "rich/PhotonDetectorSuperModule.h"

// ------------------------------------------------------------------------
//
//   the description of the detector: everything is accessable from here
//
//@Man: Using the ChannelID
//@{
/// return the unique channel-ID=1000000*sup+10000*row+100*col+pmchan;
  inline int ChannelID(int sup,int row,int col,int pmchan) {
    return 1000000*sup+10000*row+100*col+pmchan;
    }
/// determine the supermodule-ID using the ChannelID 
  inline int smFromChannelID(int i)     { return i/1000000; }
/// determine the supermodule row using the ChannelID 
  inline int rowFromChannelID(int i)    { return (i%1000000)/10000; }
/// determine the supermodule column  using the ChannelID 
  inline int colFromChannelID(int i)    { return (i%10000)/100; }
/// determine the PMT channel number using the ChannelID 
  inline int pmchanFromChannelID(int i) { return i%100; }
//@}
/// root for the photondetector description
/**
 The Photondetector is described by the class RICHDetector.
 The detector is described as a tree structure, where the first level
 consists out of the supermodule, the second is the description of the
 detector cells, especially which kind of pmt is sitting in the cell.
 The last level describe the pmt.
 The detector consists out of many supermodules, which are described
 using the class PhotonDetectorSupermodule. The number of supermodules
 you can get using the function NumSuperModule() and the information
 of the i'th supermodule you can access using the call "SuperModule(i)".
 */
 

class RICHDetector {
List<PMType>        pmtype_list;
List<PM>            pm_list;
List<LensSystem>    lens_list;

int numsm,lasthit;
PhotonDetectorSuperModule **sm;

int numchannel,firstCEFFindex;
int s,r,c,p,maxp;
public:

/// simulate the detection process of a photon. See also page:
/**
 The function "dodetect" is called to detect a photon. The function
 goes through the tree and check if the photon is detected.
 For this it first searches the hit supermodule. After this it takes
 this supermodule and call the "dodetect"-funtion of the class
 PhotonDetectorSupermodule, which handle the further search through the 
 tree. 
 The return value indicated, if the photon is detected (return value=1)
 or not (return value=0). In the first case the channelid is return
 with the last argument.
 */
   
  int  dodetect(Photon *p,int *ChannelID);
  
/// determine intersection point with the detector plane
  void intersection(const StraightLine &p,double *path);
  int  hit_in_acceptance(const StraightLine &p);

///  return the position of the last intersection
  inline Vec3 hitpos() const {
    Vec3 h;
    if (lasthit>=0) h=SuperModule(lasthit).hit();
    return h;
    }
/// return the number of supermodules
  inline       int        NumSuperModule() const { return numsm; }
/// returns the i'th supermodule
  inline const PhotonDetectorSuperModule& SuperModule(int i) const { 
    return *sm[i]; 
    }  

/// check if cell is equiped
  inline int ChannelExist(int id) {
    s=smFromChannelID(id);
    r=rowFromChannelID(id);
    c=colFromChannelID(id);
    p=pmchanFromChannelID(id);
//    printf("channelexist %d %2d %2d %2d %2d\n",id,s,r,c,p);
    return (SuperModule(s).Cell(r,c)!=0&&
            p<SuperModule(s).Cell(r,c)->pm().numchannel());
    }
/// check if cell is equiped
  inline int ChannelExist(int ss,int rr,int cc,int pp) {
    s=ss;r=rr;c=cc;p=pp;
    int ret=(SuperModule(s).Cell(r,c)!=0);
    if (ret) ret= (p<SuperModule(s).Cell(r,c)->pm().numchannel());
    return ret;
    }
/// access to the supermodule (to be used after call of CellExist,ChannelLoop)
  inline const PhotonDetectorSuperModule& SuperModule() const {
    return SuperModule(s);
    }
/// access to the cell (to be used after call of CellExist,ChannelLoop)
  inline PhotonDetectorCell& Cell() {
    return *SuperModule(s).Cell(r,c);
    }
/// access to the pm (to be used after call of CellExist,ChannelLoop)
  inline  PM& pm()  {
    return SuperModule(s).Cell(r,c)->pm();
    }
  inline const PM& pm() const {
    return SuperModule(s).Cell(r,c)->pm();
    }
/// returns the pmt-channel (to be used after call of CellExist,ChannelLoop)
  inline int pmChannel() const { return p; }
/// best-guess position as function of the channel ID. See also page:
/**   
 The function "pos" inverses the detection process. From a given channel number
 it determines the best-guess position of the hit on the front of the
 supermodule/lense system (to be used after call of CellExist,ChannelLoop)
 in the HERA-B coordinate system.
 */
  inline Vec3 pos() const { return SuperModule(s).pos(r,c,p);}
/// return pos() transformed to the supermodule coordinates
  inline Vec3 pos_smcoord() const {
    return SuperModule(s).TransformFromOld(pos());
    }
/// index of each channel in an array, all elements are used. (to be used after call of CellExist,ChannelLoop)
  inline int ChannelIndex()  {
    return pm().CEFFindex(p)-firstCEFFindex;
    }

/// access to an existing cell
  inline const PhotonDetectorCell& ExistingCell(int sup,int row,int col) const {
    PhotonDetectorCell* c=SuperModule(sup).Cell(row,col);
    if (!c) {
      fprintf(stderr,"Access to not existing cell: %d %d %d\n",sup,row,col);
      exit(1);
      }
    return *c;
    }

/// access to an existing pmt
  inline const PM& ExistingPM(int sup,int row,int col) const {
    return ExistingCell(sup,row,col).pm();
    }

/// total number of channels of the detector
  inline int NumChannel() { return numchannel; }

/// index of each channel in an array, all elements are used.
  inline int ChannelIndex(int s,int r,int c,int p) {
    return ExistingPM(s,r,c).CEFFindex(p)-firstCEFFindex;
    }
/// initialize loop over all existing channels
  void NewChannelLoop();

/// loop over all existing channels
  int ChannelLoop();
/// loop over all existing channels returning the specification of the channel.
  int ChannelLoop(int *sup,int *row,int *col,int *pmtchannel);
/// return if channel is dead
  inline int ChannelIsDead() const { return pm().efficiency(p)<0; }
/// return if channel is dead
  inline int ChannelIsHot() const { return pm().efficiency(p)>1; }
/// return if the channel is working
  inline int ChannelWorks() const { return !(ChannelIsHot()||ChannelIsDead()); }
/// return the channel dependend efficiency factor
  inline double efficiency() const { return pm().efficiency(p); }
/// set the channel dependend efficiency factor
  inline float& efficiency()  { return pm().efficiency(p); }
/// returns the maximal efficiency (for photon cut in gean simulation)
  void MaxEfficiency(Table& eff,int numpoint,int mode);
  
///  print the complete detector description to a set of ASCII-Files. 
  void print();

/// constructor reading information from an ASCII-File produced using 'print'. 
  RICHDetector(char* path=0);

/// destructor
  ~RICHDetector();

  };

extern RICHDetector* RICH_Detector;
#endif
